# Copyright TU Graz, Austria <sujoy.sinharoy@iaik.tugraz.at>.
#
# This file is part of Kavach <https://extgit.iaik.tugraz.at/sesys/kavach_artifacts>.
#
# Kavach is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 3 as published by the Free
# Software Foundation.
#
# Kavach is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Kavach. If not, see <https://www.gnu.org/licenses/gpl-3.0.txt>.

#!/usr/bin/env python3

"""
A script to get the FPGA memory dump
on the FPGA connecting via the ChipWhisperer API
"""

import time
import array
import logging
import argparse
import chipwhisperer as cw

LOGGER = logging.getLogger(__name__)

def setup(loglevel, logfmt):
    logging.basicConfig(level=loglevel, format=logfmt)
    LOGGER.setLevel(loglevel)


def fetch_target():
    return cw.target(None, cw.targets.CW305, fpga_id="100t", force=False)


def print_state(state, stop=2047):
    """Print state as hexadecimal values"""
    print('<state>')
    tailzero_index = None
    for i in range(stop, -1, -1):
        if state['bram'][i] != 0:
            tailzero_index = i
            break
    if tailzero_index is not None:
        while tailzero_index % 8 != 0:
            tailzero_index += 1
        if tailzero_index < 32:
            tailzero_index = 32

    for i in range(0, stop, 8):
        print(' {:04X}:'.format(i), end='')
        if i == tailzero_index:
            print(' … only zeroes follow here.')
            break
        for j in range(8):
            word = state['bram'][i + j]
            print("  {:016X}".format(word), end='')
        print()
    LOGGER.info(" status = {:X}".format(state['bram'][2047]))
    print('</state>')


def run_read(target, state, stop=2047):
    """Fetch all values of BRAM and print representation"""
    print()

    # tells FPGA to stop operation
    target.fpga_read((1 << 3) | 7, 1)

    time.sleep(2)

    target.fpga_read(1, 1)  # reset BRAM

    for i in range(stop):
        word = 0
        for j in range(8):
            # switch to USB clock
            byte = int(target.fpga_read(5, 1)[0])
            word = word + (byte << (j * 8))

            # this sequence starts and stops the increment
            target.fpga_read(2, 1)
            target.fpga_read(3, 1)
        state['bram'][i] = word

    print_state(state, stop=stop)
    print()
    return state


def main():
    """Main routine"""
    target = fetch_target()
    state = {'bram': array.array('Q', [0] * 2048)}

    state = run_read(target, state)

    LOGGER.info('Teardown finished.')


if __name__ == "__main__":
    loglevels = "CRITICAL;DEBUG;ERROR;FATAL;INFO;NOTSET;WARN;WARNING".split(";")

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "--log-level",
        dest="loglevel",
        default="DEBUG",
        choices=loglevels,
        help="loglevel for the logging module",
    )
    parser.add_argument(
        "--log-format",
        dest="logformat",
        default="__ %(asctime)s,%(levelname)s: %(message)s",
        help="log message for the logging module",
    )

    args = parser.parse_args()
    setup(loglevel=getattr(logging, args.loglevel), logfmt=args.logformat)
    main()
