# Copyright TU Graz, Austria <sujoy.sinharoy@iaik.tugraz.at>.
#
# This file is part of Kavach <https://extgit.iaik.tugraz.at/sesys/kavach_artifacts>.
#
# Kavach is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 3 as published by the Free
# Software Foundation.
#
# Kavach is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Kavach. If not, see <https://www.gnu.org/licenses/gpl-3.0.txt>.

#!/usr/bin/env python3

"""
A script to starting the cryptographic operation
on the FPGA connecting via the ChipWhisperer API.

The cryptographic operation is a masked
AOverflowBit and Polynomila vector multiplication implementation running in a
loop taking pseudo-random values as input.
"""

import time
import array
import random
import logging
import argparse
import chipwhisperer as cw


LOGGER = logging.getLogger(__name__)
STATIC_VALUE = 42


def setup(loglevel, logfmt):
    logging.basicConfig(level=loglevel, format=logfmt)
    LOGGER.setLevel(loglevel)


def setup_target(crypto_clk_freq=10e6, bitstream_file=""):
    """Set up a target acc. to CW305 python API"""
    # https://github.com/newaetech/chipwhisperer/blob/master/software/chipwhisperer/capture/targets/CW305.py#L61
    target = cw.target(None, cw.targets.CW305, bsfile=bitstream_file, fpga_id="100t", force=False)
    target.vccint_set(1.0)

    # we only need PLL1:
    target.pll.pll_enable_set(True)
    target.pll.pll_outenable_set(True, 0)
    target.pll.pll_outenable_set(True, 1)
    target.pll.pll_outenable_set(False, 2)

    # crypto clock frequency (2nd arg can be 1 or 2)
    target.pll.pll_outfreq_set(10e6, 0)
    target.pll.pll_outfreq_set(crypto_clk_freq, 1)

    target.clkusbautooff = True
    target.clksleeptime = 1

    return target

def run_sampler(target):
    #Start the operations here
    
    # switch to usb clock
    target.fpga_read(5, 1)

    # <init>
    # reset BRAM (BRAM index to 0)
    target.fpga_read(1, 1)

    # write fixed values to BRAM
    state = array.array('Q', [random.randint(0, 2**64-1) for _ in range(500)])

    for i in range(500):
        one_word = state[i]

        for _ in range(8):
            byte = one_word & 0xFF
            one_word = (one_word - byte) >> 8
            target.fpga_write(0 << 4, [byte])
            target.fpga_read(2, 1)
            target.fpga_read(3, 1)
    # </init>

    # reset BRAM (BRAM index to 0)
    target.fpga_read(1, 1)

    # set command_we0 to 1 and command_in to 0 (resetting control signals like 'done')
    for j in range(8):
        target.fpga_write(11 << 4, [0])
        target.fpga_read(2, 1)
        target.fpga_read(3, 1)

    # reset BRAM (BRAM index to 0)
    target.fpga_read(1, 1)

    # indicate that we want to run multiple instructions in a loop
    target.fpga_read(6, 1)
    
    # <trivium>
    INS = 11   # Instruction code
    OP1 = 0    # Operand 1
    OP2 = 310  # Operand 2
    WT = 0     # Write destination for result
    command = INS + (OP1<<5) + (OP2<<16) + (WT<<27)
    # command_we0 is set as 1 and command is sent (byte by byte)
    # wea_ext_core is set to 0 + wea_ext_ISA is set to 1 (useless)
    one_word = command
    for _ in range(8):
        byte = one_word & 0xFF
        one_word = (one_word - byte) >> 8
        target.fpga_write(9 << 4, [byte])
        target.fpga_read(2, 1)  # increases byte address by one
        target.fpga_read(3, 1)  # stops increasing byte address by one
    # </trivium>
    
    # <wait> 
    INS = 29       # Instruction code
    OP1 = 2      # Operand 1
    OP2 = 0      # Operand 2
    WT = 0      # Write destination for result
    command = INS + (OP1<<5) + (OP2<<16) + (WT<<27)
    one_word = command
    for j in range(8):
        byte = one_word & 0xFF
        one_word = (one_word - byte) >> 8
        register = 9 << 4
        target.fpga_write(register, [byte])
        target.fpga_read(2, 1)
        target.fpga_read(3, 1)
    # </wait>
    
    # <sub>
    # this is done to mask the secret
    xorpoly_iLen = 48  # bytes
    command_xorpoly_params = xorpoly_iLen 
    one_word = command_xorpoly_params
    for j in range(8):
        byte = one_word & 0xFF
        one_word = (one_word - byte) >> 8
        target.fpga_write(17 << 4, [byte])
        target.fpga_read(2, 1)
        target.fpga_read(3, 1)  
            
    INS = 25      # Instruction code
    OP1 = 360     # Operand 1
    OP2 = 210     # Operand 2
    WT = 310      # Write destination for result
    command = INS + (OP1<<5) + (OP2<<16) + (WT<<27)
    one_word = command
    for j in range(8):
        byte = one_word & 0xFF
        one_word = (one_word - byte) >> 8
        register = 9 << 4
        target.fpga_write(register, [byte])
        target.fpga_read(2, 1)
        target.fpga_read(3, 1)
    # </sub> 
    
    
    # <set-random>
    # this is to set the gamma variable for AOverflowBit
    INS = 28   # Instruction code
    OP1 = 0    # Operand 1
    OP2 = 0    # Operand 2
    WT = 0     # Write destination for result
    command = INS + (OP1<<5) + (OP2<<16) + (WT<<27)
    one_word = command
    for _ in range(8):
        byte = one_word & 0xFF
        one_word = (one_word - byte) >> 8
        target.fpga_write(9 << 4, [byte])
        target.fpga_read(2, 1)  # increases byte address by one
        target.fpga_read(3, 1)  # stops increasing byte address by one
    # </set-random>

    # ----------------------------------------------------------------------------------

    # <start-trigger>
    INS = 26   # Instruction code
    OP1 = 0    # Operand 1
    OP2 = 0    # Operand 2
    WT = 0     # Write destination for result
    command = INS + (OP1<<5) + (OP2<<16) + (WT<<27)
    one_word = command
    for _ in range(8):
        byte = one_word & 0xFF
        one_word = (one_word - byte) >> 8
        target.fpga_write(9 << 4, [byte])
        target.fpga_read(2, 1)  # increases byte address by one
        target.fpga_read(3, 1)  # stops increasing byte address by one
    # </start-trigger>
    
    # <wait> 
    # wait for trigger noise to die down
    INS = 29       # Instruction code
    OP1 = 2000      # Operand 1
    OP2 = 0      # Operand 2
    WT = 0      # Write destination for result
    command = INS + (OP1<<5) + (OP2<<16) + (WT<<27)
    one_word = command
    for j in range(8):
        byte = one_word & 0xFF
        one_word = (one_word - byte) >> 8
        register = 9 << 4
        target.fpga_write(register, [byte])
        target.fpga_read(2, 1)
        target.fpga_read(3, 1)
    # </wait>

    # ----------------------------------------------------------------------------------
    
     # ----------------------------------------------------------------------------------
    
    # <a2a>
    INS = 20   # Instruction code
    OP1 = 210  # Operand 1 secret as 160,260
    OP2 = 0    # Operand 2
    WT = 410   # Write destination for result
    command = INS + (OP1<<5) + (OP2<<16) + (WT<<27)
    one_word = command
    for _ in range(8):
        byte = one_word & 0xFF
        one_word = (one_word - byte) >> 8
        target.fpga_write(9 << 4, [byte])
        target.fpga_read(2, 1)  # increases byte address by one
        target.fpga_read(3, 1)  # stops increasing byte address by one
    # </a2a>

    # <vmul>
    INS = 21   # Instruction code
    OP1 = 2    # Operand 1
    OP2 = 410  # Operand 2
    WT = 568   # Write destination for result
    command = INS + (OP1<<5) + (OP2<<16) + (WT<<27)
    one_word = command
    for _ in range(8):
        byte = one_word & 0xFF
        one_word = (one_word - byte) >> 8
        target.fpga_write(9 << 4, [byte])
        target.fpga_read(2, 1)  # increases byte address by one
        target.fpga_read(3, 1)  # stops increasing byte address by one
    # </vmul>
   
     # <vmul>
    INS = 6    # Instruction code
    OP1 = 2    # Operand 1
    OP2 = 210  # Operand 2
    WT = 440   # Write destination for result
    command = INS + (OP1<<5) + (OP2<<16) + (WT<<27)
    one_word = command
    for _ in range(8):
        byte = one_word & 0xFF
        one_word = (one_word - byte) >> 8
        target.fpga_write(9 << 4, [byte])
        target.fpga_read(2, 1)  # increases byte address by one
        target.fpga_read(3, 1)  # stops increasing byte address by one
    # </vmul>

    # <vmul>
    INS = 22   # Instruction code
    OP1 = 2    # Operand 1
    OP2 = 310  # Operand 2
    WT = 504   # Write destination for result
    command = INS + (OP1<<5) + (OP2<<16) + (WT<<27)
    one_word = command
    for _ in range(8):
        byte = one_word & 0xFF
        one_word = (one_word - byte) >> 8
        target.fpga_write(9 << 4, [byte])
        target.fpga_read(2, 1)  # increases byte address by one
        target.fpga_read(3, 1)  # stops increasing byte address by one
    # </vmul>
   
    # ----------------------------------------------------------------------------------

    # ----------------------------------------------------------------------------------
    

     # <wait> 
    INS = 29       # Instruction code
    OP1 = 2000      # Operand 1
    OP2 = 0      # Operand 2
    WT = 0      # Write destination for result
    command = INS + (OP1<<5) + (OP2<<16) + (WT<<27)
    one_word = command
    for j in range(8):
        byte = one_word & 0xFF
        one_word = (one_word - byte) >> 8
        register = 9 << 4
        target.fpga_write(register, [byte])
        target.fpga_read(2, 1)
        target.fpga_read(3, 1)
    # </wait>
    
    # <wait> 
    INS = 29       # Instruction code
    OP1 = 2000      # Operand 1
    OP2 = 0      # Operand 2
    WT = 0      # Write destination for result
    command = INS + (OP1<<5) + (OP2<<16) + (WT<<27)
    one_word = command
    for j in range(8):
        byte = one_word & 0xFF
        one_word = (one_word - byte) >> 8
        register = 9 << 4
        target.fpga_write(register, [byte])
        target.fpga_read(2, 1)
        target.fpga_read(3, 1)
    # </wait>
   
     # <trivium>
    INS = 11   # Instruction code
    OP1 = 0    # Operand 1
    OP2 = 1    # Operand 2
    WT = 0     # Write destination for result
    command = INS + (OP1<<5) + (OP2<<16) + (WT<<27)
    one_word = command
    for _ in range(8):
        byte = one_word & 0xFF
        one_word = (one_word - byte) >> 8
        target.fpga_write(9 << 4, [byte])
        target.fpga_read(2, 1)  # increases byte address by one
        target.fpga_read(3, 1)  # stops increasing byte address by one
    # </trivium>
     # <wait> 
    INS = 29       # Instruction code
    OP1 = 1      # Operand 1
    OP2 = 0      # Operand 2
    WT = 0      # Write destination for result
    command = INS + (OP1<<5) + (OP2<<16) + (WT<<27)
    one_word = command
    for j in range(8):
        byte = one_word & 0xFF
        one_word = (one_word - byte) >> 8
        register = 9 << 4
        target.fpga_write(register, [byte])
        target.fpga_read(2, 1)
        target.fpga_read(3, 1)
    # </wait>
    
    
    # <copy> 
    INS = 12       # Instruction code
    OP1 = 160      # Operand 1
    OP2 = 48      # Operand 2
    WT = 210      # Write destination for result
    command = INS + (OP1<<5) + (OP2<<16) + (WT<<27)
    one_word = command
    for j in range(8):
        byte = one_word & 0xFF
        one_word = (one_word - byte) >> 8
        register = 9 << 4
        target.fpga_write(register, [byte])
        target.fpga_read(2, 1)
        target.fpga_read(3, 1)
    # </copy>
    # <copy> 
    INS = 12       # Instruction code
    OP1 = 260      # Operand 1
    OP2 = 48      # Operand 2
    WT = 310      # Write destination for result
    command = INS + (OP1<<5) + (OP2<<16) + (WT<<27)
    one_word = command
    for j in range(8):
        byte = one_word & 0xFF
        one_word = (one_word - byte) >> 8
        register = 9 << 4
        target.fpga_write(register, [byte])
        target.fpga_read(2, 1)
        target.fpga_read(3, 1)
    # </copy>
    
    # <wait> 
    INS = 29       # Instruction code
    OP1 = 2000      # Operand 1
    OP2 = 0      # Operand 2
    WT = 0      # Write destination for result
    command = INS + (OP1<<5) + (OP2<<16) + (WT<<27)
    one_word = command
    for j in range(8):
        byte = one_word & 0xFF
        one_word = (one_word - byte) >> 8
        register = 9 << 4
        target.fpga_write(register, [byte])
        target.fpga_read(2, 1)
        target.fpga_read(3, 1)
    # </wait>
     # <wait> 
    INS = 29       # Instruction code
    OP1 = 2000      # Operand 1
    OP2 = 0      # Operand 2
    WT = 0      # Write destination for result
    command = INS + (OP1<<5) + (OP2<<16) + (WT<<27)
    one_word = command
    for j in range(8):
        byte = one_word & 0xFF
        one_word = (one_word - byte) >> 8
        register = 9 << 4
        target.fpga_write(register, [byte])
        target.fpga_read(2, 1)
        target.fpga_read(3, 1)
    # </wait>
    
     # <wait> 
    INS = 29       # Instruction code
    OP1 = 2000      # Operand 1
    OP2 = 0      # Operand 2
    WT = 0      # Write destination for result
    command = INS + (OP1<<5) + (OP2<<16) + (WT<<27)
    one_word = command
    for j in range(8):
        byte = one_word & 0xFF
        one_word = (one_word - byte) >> 8
        register = 9 << 4
        target.fpga_write(register, [byte])
        target.fpga_read(2, 1)
        target.fpga_read(3, 1)
    # </wait>
   
    
    
      # <a2a>
    INS = 20   # Instruction code
    OP1 = 210  # Operand 1 secret as 160,260
    OP2 = 0    # Operand 2
    WT = 410   # Write destination for result
    command = INS + (OP1<<5) + (OP2<<16) + (WT<<27)
    one_word = command
    for _ in range(8):
        byte = one_word & 0xFF
        one_word = (one_word - byte) >> 8
        target.fpga_write(9 << 4, [byte])
        target.fpga_read(2, 1)  # increases byte address by one
        target.fpga_read(3, 1)  # stops increasing byte address by one
    # </a2a>

    # <vmul>
    INS = 21   # Instruction code
    OP1 = 2    # Operand 1
    OP2 = 410  # Operand 2
    WT = 568   # Write destination for result
    command = INS + (OP1<<5) + (OP2<<16) + (WT<<27)
    one_word = command
    for _ in range(8):
        byte = one_word & 0xFF
        one_word = (one_word - byte) >> 8
        target.fpga_write(9 << 4, [byte])
        target.fpga_read(2, 1)  # increases byte address by one
        target.fpga_read(3, 1)  # stops increasing byte address by one
    # </vmul>
   
    
     # <vmul>
    INS = 6    # Instruction code
    OP1 = 2    # Operand 1
    OP2 = 210  # Operand 2
    WT = 440   # Write destination for result
    command = INS + (OP1<<5) + (OP2<<16) + (WT<<27)
    one_word = command
    for _ in range(8):
        byte = one_word & 0xFF
        one_word = (one_word - byte) >> 8
        target.fpga_write(9 << 4, [byte])
        target.fpga_read(2, 1)  # increases byte address by one
        target.fpga_read(3, 1)  # stops increasing byte address by one
    # </vmul>
   
    # <vmul>
    INS = 22   # Instruction code
    OP1 = 2    # Operand 1
    OP2 = 310  # Operand 2
    WT = 504   # Write destination for result
    command = INS + (OP1<<5) + (OP2<<16) + (WT<<27)
    one_word = command
    for _ in range(8):
        byte = one_word & 0xFF
        one_word = (one_word - byte) >> 8
        target.fpga_write(9 << 4, [byte])
        target.fpga_read(2, 1)  # increases byte address by one
        target.fpga_read(3, 1)  # stops increasing byte address by one
    # </vmul>

    # <wait> 
    INS = 29       # Instruction code
    OP1 = 100      # Operand 1
    OP2 = 0      # Operand 2
    WT = 0      # Write destination for result
    command = INS + (OP1<<5) + (OP2<<16) + (WT<<27)
    one_word = command
    for j in range(8):
        byte = one_word & 0xFF
        one_word = (one_word - byte) >> 8
        register = 9 << 4
        target.fpga_write(register, [byte])
        target.fpga_read(2, 1)
        target.fpga_read(3, 1)
    # </wait>
    

   
    # ----------------------------------------------------------------------------------

    # <end-trigger>
    INS = 27   # Instruction code
    OP1 = 0    # Operand 1
    OP2 = 0    # Operand 2
    WT = 0     # Write destination for result
    command = INS + (OP1<<5) + (OP2<<16) + (WT<<27)
    one_word = command
    for _ in range(8):
        byte = one_word & 0xFF
        one_word = (one_word - byte) >> 8
        target.fpga_write(9 << 4, [byte])
        target.fpga_read(2, 1)  # increases byte address by one
        target.fpga_read(3, 1)  # stops increasing byte address by one
    # </end-trigger>

    # ----------------------------------------------------------------------------------

    time.sleep(0.5)

    # indicate that we finished writing multiple instructions
    target.fpga_read(7, 1)

   
def main(crypto_clk_freq, bitstream_file=""):
    target = setup_target(crypto_clk_freq=crypto_clk_freq, bitstream_file=bitstream_file)
    run_sampler(target)
    LOGGER.info('Setup finished. Cryptographic operation is running …')


if __name__ == '__main__':
    loglevels = 'CRITICAL;DEBUG;ERROR;FATAL;INFO;NOTSET;WARN;WARNING'.split(';')

    parser = argparse.ArgumentParser(description=__doc__.strip())
    parser.add_argument('--bitstream-file', help='the bitstream file to flash to device before setup')
    parser.add_argument('--clock-frequency', type=int, default=10000000, help='PLL crypto clock frequency to be configured at CW305 in cycles per second')
    parser.add_argument('--log-level', dest='loglevel', default='DEBUG', choices=loglevels, help='loglevel for the logging module')
    parser.add_argument('--log-format', dest='logformat', default='%(asctime)s,%(levelname)s: %(message)s', help='log message for the logging module')

    args = parser.parse_args()
    setup(loglevel=getattr(logging, args.loglevel), logfmt=args.logformat)
    main(args.clock_frequency, args.bitstream_file)
