# Copyright TU Graz, Austria <sujoy.sinharoy@iaik.tugraz.at>.
#
# This file is part of Kavach <https://extgit.iaik.tugraz.at/sesys/kavach_artifacts>.
#
# Kavach is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 3 as published by the Free
# Software Foundation.
#
# Kavach is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Kavach. If not, see <https://www.gnu.org/licenses/gpl-3.0.txt>.

#!/usr/bin/env python3

"""
Take measurements for the Kavach implementation
on a CW305 with PicoScope 6000.
"""

import os
import sys
import time
import struct
import ctypes
import typing
import logging
import pathlib
import hashlib
import datetime
import argparse
import statistics
import subprocess
import numpy as np
from picosdk.ps6000 import ps6000 as ps
from picosdk.functions import adc2mV, assert_pico_ok, mV2adc

LOGGER = logging.getLogger(__name__)

# admissible values: via https://github.com/picotech/picosdk-python-wrappers/blob/e850aa6/picosdk/ps6000.py#L33
MEASUREMENT_RANGE_CRYPTO = "PS6000_1V"
MEASUREMENT_RANGE_TRIGGER = "PS6000_500MV"


def setup(logfile: str, loglevel: int, logfmt: str):
    # generic, but unused config
    logging.basicConfig(level=loglevel, format=logfmt)

    # silence the verbose matplotlib
    logging.getLogger('matplotlib').setLevel(logging.ERROR)

    # actual configuration
    LOGGER.setLevel(loglevel)
    formatter = logging.Formatter(logfmt)

    #streamhandler = logging.StreamHandler()
    #streamhandler.setLevel(logging.INFO)
    #streamhandler.setFormatter(formatter)

    # also write to a file
    filehandler = logging.FileHandler(logfile, mode="a")
    filehandler.setLevel(loglevel)
    filehandler.setFormatter(formatter)

    LOGGER.addHandler(filehandler)
    #LOGGER.addHandler(streamhandler)


def timebase_to_sample_interval(timebase: int) -> int:
    """Given a picosdk timebase, return the number of nanoseconds between two samples"""
    assert 0 <= timebase <= 2**32 - 1
    if 0 <= timebase <= 4:
        return ((2**timebase)/5)
    else:
        return ((timebase - 4) * 6.4)


def human_readable_hertz(ops_per_second: int) -> str:
    """Given the number of operations per second,
    return the Hertz in a human-readable way
    """
    for unit in ['','K','M','G','T','P','E','Z']:
        if abs(ops_per_second) < 1000.0:
            return "%3.1f%s%s" % (ops_per_second, unit, 'Hz')
        ops_per_second /= 1000.0
    return "%.1f%s%s" % (ops_per_second, 'Y', 'Hz')


def sha256_digest(filepath: str) -> str:
    """Compute SHA256 digest of a file and returns its hexadecimal digest"""
    h = hashlib.new("sha256")
    with open(filepath, 'rb') as fd:
        while True:
            block = fd.read(4096)
            h.update(block)
            if not block:
                break
    return h.hexdigest()


def setup_cryptographic_operation(python3: str=None, setup_script: str='crypto_setup_unprotected_static_binomial_sampler.py', clock_frequency: int=10000000, bitstream_file: str=''):
    """Start the cryptographic operation"""
    if python3 is None:
        python3 = os.getenv('PYTHON3')
    if python3 is None:
        python3 = 'python3'

    assert python3
    assert setup_script
    cmdline = [python3, setup_script, '--clock-frequency', str(clock_frequency)]
    if bitstream_file:
        cmdline.append("--bitstream-file")
        cmdline.append(bitstream_file)
    LOGGER.info("executing {}".format(' '.join(cmdline)))

    proc = subprocess.Popen(cmdline,
        stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    stdout, stderr = proc.communicate()
    if stdout:
        print(f'<stdout>\n{stdout.decode("utf-8")}</stdout>')
    if stderr:
        print(f'<stderr>\n{stderr.decode("utf-8")}</stderr>', file=sys.stderr)
    if proc.returncode == 0:
        LOGGER.debug(f'setup script exited with {proc.returncode}')
    else:
        raise ValueError(f'setup script exited with {proc.returncode}')


def write_mbin_file(matrix, filepath):
    """Write matrix data to filepath"""
    if isinstance(matrix, np.ndarray):
        # NOTE for compatibility with an older version,
        # we accept numpy ndim arrays, but in this case we need to check the data type
        assert matrix.dtype == np.float16
    height = len(matrix)
    width = len(matrix[0])
    assert 0 <= width < 2**32
    assert 0 <= height < 2**32

    with open(filepath, 'wb') as fd:
        # write header
        fd.write(struct.pack('<bbbbbbbbbbbbIIIII', *bytes(b'matrix-f16  '), 0, 0, 0, height, width))
        # 1. mandatory 10 bytes 'matrix-f16'
        # 2. 14 non-standardized/arbitrary bytes
        # 3. 4 little-endian bytes specifying number of rows
        # 4. 4 little-endian bytes specifying number of columns
        LOGGER.debug("rows = {}  columns = {}".format(height, width))

        # write entries
        for r in range(height):
            for c in range(width):
                # 5. f16 value per entry
                item = float(matrix[r][c])
                fd.write(struct.pack('<e', item))


def start_picoscope() -> dict:
    """Start the picoscope and get a handle"""
    handle = ctypes.c_int16()
    state = {'handle': handle}

    # Open 6000 series PicoScope
    # Returns handle to handle for use in future API functions
    LOGGER.info('open device handle')
    state["openunit"] = ps.ps6000OpenUnit(ctypes.byref(handle), None)
    assert_pico_ok(state["openunit"])

    # Set up channel A → power consumption during cryptographic operation
    #   handle = state['handle']
    #   channel = PS6000_CHANNEL_A = 0
    #   enabled = 1
    #   coupling type = PS6000_DC = 1
    #   range = PS6000_2V = 7
    #   analogue offset = 0 V
    #   bandwidth limiter = PS6000_BW_FULL = 0
    LOGGER.info('set up channel A')
    channel_a = ps.PS6000_CHANNEL["PS6000_CHANNEL_A"]
    channel_a_enabled = True
    # 0 PS6000_AC, 1 MΩ impedance, AC coupling
    # 1 PS6000_DC_1M, 1 MΩ impedance, DC coupling
    # 2 PS6000_DC_50R, DC coupling, 50 Ω impedance
    channel_a_coupling_type = 2 # try 2
    state['channel_a_range'] = ps.PS6000_RANGE[MEASUREMENT_RANGE_CRYPTO]
    state["setChA"] = ps.ps6000SetChannel(
        state['handle'],
        channel_a,
        channel_a_enabled,
        channel_a_coupling_type,
        state['channel_a_range'],0,0
    )
    assert_pico_ok(state["setChA"])

    # Set up channel B → trigger signal
    #   handle = state['handle']
    #   channel = PS6000_CHANNEL_B = 1
    #   enabled = 1
    #   coupling type = PS6000_DC = 1
    #   range = PS6000_2V = 7, PS6000_5V = 8
    #   analogue offset = 0 V
    #   bandwidth limiter = PS6000_BW_FULL = 0
    LOGGER.info('set up channel B')
    channel_b = ps.PS6000_CHANNEL["PS6000_CHANNEL_B"]
    channel_b_enabled = True
    channel_b_coupling_type = True  # True = DC, False = AC
    state['channel_b_range'] = ps.PS6000_RANGE[MEASUREMENT_RANGE_TRIGGER]
    state["setChB"] = ps.ps6000SetChannel(
        state['handle'],
        channel_b,
        channel_b_enabled,
        channel_b_coupling_type,
        state['channel_b_range'],0,0
    )
    assert_pico_ok(state["setChB"])
    
    # 2 == PS_6000_FREQUENCY_10MHz
    # 30000 is threshold voltage, a bit less than 1V
    state["clock"] = ps.ps6000SetExternalClock(state['handle'], 2, 30000) 
    assert_pico_ok(state["clock"])
    
    #   enabled = 1
    #   source = PS6000_CHANNEL_A = 0
    #   threshold = 64 ADC counts
    #   direction = PS6000_RISING = 2
    #   delay = 0 s
    #   auto Trigger = 1000 ms

    # Set up single trigger
    #   handle = state['handle']
    #   enabled = 1
    #   source = PS6000_CHANNEL_A = 0
    #   threshold = 64 ADC counts
    #   direction = PS6000_RISING = 2
    #   delay = 0 s
    #   auto Trigger = 1000 ms
    # mV2adc = takes a voltage value and converts it into adc counts
    LOGGER.info('set up trigger')
    max_adc = ctypes.c_int16(32512)
    source = ps.PS6000_CHANNEL["PS6000_CHANNEL_B"]
    threshold = int(mV2adc(100, state['channel_b_range'], max_adc))
    trigger_direction =ps.PS6000_THRESHOLD_DIRECTION["PS6000_RISING"]# 2  # = PS6000_RISING
    trigger_delay = 0  # seconds
    state['trigger_auto_trigger'] = 10_000  # ms
    state["trigger"] = ps.ps6000SetSimpleTrigger(
        state['handle'],
        1,
        source,
        threshold,
        trigger_direction,
        trigger_delay,
        state['trigger_auto_trigger'],
    )
    assert_pico_ok(state["trigger"])

    return state


def run_block_with_picoscope(state, timebase=8, number_of_samples=20_000_000):
    # Set number of pre and post trigger samples to be collected
    # FTR: 32 MS seems to mean 0 <= number_of_samples <= 16_000_000
    pre_trigger_samples = 500
    post_trigger_samples = number_of_samples - 500
    max_samples = pre_trigger_samples + post_trigger_samples

    # Get timebase information
    #   handle = state['handle']
    #   noSamples = max_samples
    #   pointer to timeIntervalNanoseconds = ctypes.byref(time_intervalns)
    #   pointer to max_samples = ctypes.byref(returned_max_samples)
    # Get timebase information
    # handle = state['handle']
    timebase = timebase
    timebase_oversample = 0
    timebase_interval_ns = ctypes.c_float()
    #print(timebase_interval_ns.value)
    timebase_segment_index = 0
    state["getTimebase"] = ps.ps6000GetTimebase2(
        state['handle'],
        timebase,
        max_samples,
        ctypes.byref(timebase_interval_ns),
        timebase_oversample,
        None,
        timebase_segment_index,
    )
    #print(timebase_interval_ns.value)
    assert_pico_ok(state["getTimebase"])

    # Run block capture
    #   handle = state['handle']
    #   number of pre-trigger samples = pre_trigger_samples
    #   number of post-trigger samples = PostTriggerSamples
    #   timebase = 8 = 80 ns (see Programmer's guide for mre information on timebases)
    #   oversample = 0
    #   time indisposed ms = None (not needed in the example)
    #   segment index = 0
    #   lpReady = None (using ps6000IsReady rather than ps6000BlockReady)
    #   pParameter = None
    state["runBlock"] = ps.ps6000RunBlock(
        state['handle'],
        pre_trigger_samples,
        post_trigger_samples,
        timebase,
        timebase_oversample,
        None,
        0,
        None,
        None,
    )
    assert_pico_ok(state["runBlock"])

    LOGGER.info("Waiting for trigger …")

    # Check for data collection to finish using ps6000IsReady
    ready = ctypes.c_int16(0)
    check = ctypes.c_int16(0)
    start = time.time()
    while ready.value == check.value:
        state["isReady"] = ps.ps6000IsReady(state['handle'], ctypes.byref(ready))

    actual_duration = time.time() - start
    dont_exceed_duration = min(0.95 * state['trigger_auto_trigger'], state['trigger_auto_trigger'] - 1000.) / 1000
    if actual_duration > dont_exceed_duration:
        LOGGER.warning('waiting time ({:.1f}s) close to timeout ({}s) - did it trigger?'.format(actual_duration, state['trigger_auto_trigger'] / 1000.))
    LOGGER.info("Trigger signal received.")

    # Create buffers ready for assigning pointers for data collection
    buffer_a_max = (ctypes.c_int16 * max_samples)()
    buffer_a_min = (
        ctypes.c_int16 * max_samples
    )()  # used for downsampling which isn't in the scope of this example
    buffer_b_max = (ctypes.c_int16 * max_samples)()
    buffer_b_min = (
        ctypes.c_int16 * max_samples
    )()  # used for downsampling which isn't in the scope of this example

    # Set data buffer location for data collection from channel A
    # handle = state['handle']
    # source = PS6000_CHANNEL_A = 0
    # pointer to buffer max = ctypes.byref(buffer_a_max)
    # pointer to buffer min = ctypes.byref(buffer_a_min)
    # buffer length = max_samples
    # ratio mode = PS6000_RATIO_MODE_NONE = 0
    
    state["setDataBuffersA"] = ps.ps6000SetDataBufferBulk(
        state['handle'],
        ps.PS6000_CHANNEL["PS6000_CHANNEL_A"], 
        buffer_a_max,
         max_samples, 
         0, 
         0)
            

       
    assert_pico_ok(state["setDataBuffersA"])

    # Set data buffer location for data collection from channel B
    # handle = state['handle']
    # source = PS6000_CHANNEL_B = 1
    # pointer to buffer max = ctypes.byref(buffer_b_max)
    # pointer to buffer min = ctypes.byref(buffer_b_min)
    # buffer length = max_samples
    # ratio mode = PS6000_RATIO_MODE_NONE = 0
    state["setDataBuffersB"] = ps.ps6000SetDataBufferBulk(
        state['handle'],
        ps.PS6000_CHANNEL["PS6000_CHANNEL_B"], 
        buffer_b_max,
         max_samples, 
         0, 
         0
    )
    assert_pico_ok(state["setDataBuffersB"])

    # create overflow location
    overflow = ctypes.c_int16()
    # create converted type max_samples
    cmax_samples = ctypes.c_int32(max_samples)

    # Retried data from scope to buffers assigned above
    # handle = state['handle']
    # start index = 0
    # pointer to number of samples = ctypes.byref(cmax_samples)
    # downsample ratio = 1
    # downsample ratio mode = PS6000_RATIO_MODE_NONE
    # pointer to overflow = ctypes.byref(overflow))
    state["getValues"] = ps.ps6000GetValues(
        state['handle'], 0,  ctypes.byref(cmax_samples), 1, 0, 0, ctypes.byref(overflow)
    )
    assert_pico_ok(state["getValues"])

    # find maximum ADC count value
    max_adc = ctypes.c_int16(32512)

    # convert ADC counts data to mV
    adc2mV_channel_a_max = adc2mV(buffer_a_max, state['channel_a_range'], max_adc)
    adc2mV_channel_b_max = adc2mV(buffer_b_max, state['channel_b_range'], max_adc)
    #print(adc2mV_channel_a_max)

    LOGGER.info("{} samples have been taken.".format(len(adc2mV_channel_b_max)))
    assert len(adc2mV_channel_a_max) == len(adc2mV_channel_b_max)
    assert len(adc2mV_channel_a_max) <= max_samples
    assert len(adc2mV_channel_b_max) <= max_samples

    if len(adc2mV_channel_a_max) < max_samples:
        LOGGER.info('Collected only {} samples, but expected {}'.format(len(adc2mV_channel_a_max), max_samples))
    if len(adc2mV_channel_b_max) < max_samples:
        LOGGER.info('Collected only {} samples, but expected {}'.format(len(adc2mV_channel_b_max), max_samples))

    # adc2mV_channel_a_max = […]
    # adc2mV_channel_b_max = […]
    # channel_a_range: int = 3
    # channel_b_range: int = 3
    # timebase_interval_ns = 48
    return (
        adc2mV_channel_a_max,
        adc2mV_channel_b_max,
        MEASUREMENT_RANGE_CRYPTO[7:],
        MEASUREMENT_RANGE_TRIGGER[7:],
        timebase_interval_ns.value
    )


def find_high_ranges(trigger_measurement, trigger_threshold):
    """Find all ranges in `trigger_measurement` where the trigger signal
    is set to HIGH defined by exceeding `trigger_threshold`.
    """
    trigger = trigger_measurement
    high_ranges = []
    rlen = lambda r: r[1] - r[0]

    # <determine-ranges>
    start = None
    for i in range(1, len(trigger)):
        if trigger[i - 1] < trigger_threshold and trigger[i] >= trigger_threshold:
            start = i
        elif trigger[i - 1] >= trigger_threshold and trigger[i] < trigger_threshold:
            high_ranges.append((start, i))
            start = None
    if start is not None:
        high_ranges.append((start, len(trigger)))
    # </determine-ranges>

    LOGGER.debug("{} ranges found = {}".format(len(high_ranges), ' '.join('{}–{}'.format(start, end) for start, end in high_ranges)))
    LOGGER.debug("range lengths = {}".format(' '.join(str(rlen(r)) for r in high_ranges)))

    # <potentially-drop-first-or-last>
    # remove first or/and last range if they are shorter than others
    if len(high_ranges) > 0:
        lengths_mid = [rlen(high_range) for high_range in high_ranges[1:-1]]
        low, high = min(lengths_mid), max(lengths_mid)
        if len(high_ranges) > 2:
            if not (low <= rlen(high_ranges[0]) <= high):
                LOGGER.debug('dropping the first range, because it has an unusual length')
                high_ranges.pop(0)
            if not (low <= rlen(high_ranges[-1]) <= high):
                LOGGER.debug('dropping the last range, because it has an unusual length')
                high_ranges.pop()
    # </potentially-drop-first-or-last>

    return high_ranges


def find_clusters_of_high_ranges(high_ranges):
    """Clusterize high_ranges and return (min, median, max) of each cluster"""
    clusters = []
    rlen = lambda r: r[1] - r[0]

    # <determine-cluster>
    lengths = [rlen(high_range) for high_range in high_ranges]
    unique_lengths = set(lengths)
    while unique_lengths:
        # put values a and b with abs(a - b) <= 2 into the same cluster
        val = unique_lengths.pop()
        cluster = {val,}
        for offset in range(-2, 3):
            if (val + offset) in unique_lengths:
                cluster.add(val + offset)
                unique_lengths.remove(val + offset)
        clusters.append(cluster)
    clusters = [(min(cluster), statistics.median(cluster), max(cluster)) for cluster in clusters]
    # </determine-cluster>

    LOGGER.debug("{} cluster(s): {}".format(len(clusters), ' '.join(map(lambda c: str(c[1]), clusters))))
    LOGGER.debug("  {}".format(' '.join('{}…{}'.format(start, end) for (start, _, end) in clusters)))

    return clusters


def stop_picoscope(state: dict):
    # Create time data
    state["stop"] = ps.ps6000Stop(state['handle'])
    assert_pico_ok(state["stop"])

    # Close unitDisconnect the scope
    ps.ps6000CloseUnit(state['handle'])


def teardown_cryptographic_operation(python3=None, script='crypto_teardown.py'):
    """Stop the cryptographic operation"""
    if python3 is None:
        python3 = os.getenv('PYTHON3')
    if python3 is None:
        python3 = 'python3'

    cmdline = [python3, script]
    LOGGER.info("executing {}".format(' '.join(cmdline)))

    proc = subprocess.Popen(cmdline,
        stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    stdout, stderr = proc.communicate()
    LOGGER.debug(f'teardown script exited with {proc.returncode}')
    if stdout:
        print(f'<stdout>\n{stdout.decode("utf-8")}</stdout>')
    if stderr:
        print(f'<stderr>\n{stderr.decode("utf-8")}</stderr>', file=sys.stderr)


def main(data_dir, project, bitstream_file, setup_script, teardown_script, timebase=4, clock_frequency=10000000, measurements=10, trigger_threshold=100, python_exec=None):
    """Main routine performing measurement and storing results."""
    path_project = os.path.join(data_dir, project)
    path_measurements = os.path.join(path_project, 'measurements')

    pathlib.Path(path_project).mkdir(parents=True, exist_ok=True)
    setup(logfile=os.path.join(args.data_dir, args.project, "stderr_stdout.log"), loglevel=getattr(logging, args.loglevel), logfmt=args.logformat)

    # report parameters
    LOGGER.info("Going to collect {} samples per clock cycles as part of project '{}' in '{}'".format(clock_frequency/(timebase_to_sample_interval(timebase)*100000), project, data_dir))
    if bitstream_file:
        LOGGER.info("Going to flash device with '{}' first".format(bitstream_file))
    LOGGER.info("The current UTC timestamp is {}".format(datetime.datetime.utcnow().isoformat()))
    LOGGER.info("PLL clock frequency = {}".format(human_readable_hertz(clock_frequency)))
    LOGGER.info("Number of measurements = {}".format(measurements))
    LOGGER.info("timebase = {} corresponds to {} ns between two samples".format(timebase, timebase_to_sample_interval(timebase)))
    LOGGER.debug("For example, an operation with 179 cycles should take {} samples".format(int(179 * (clock_frequency / 100_000) / timebase_to_sample_interval(timebase))))
    LOGGER.info("trigger is considered HIGH if ≥ {} mV".format(trigger_threshold))
    LOGGER.info("Script '{}' will be used for cryptographic operation setup".format(setup_script))

    # verify parameters
    if not os.path.exists(python_exec):
        raise IOError("Executable '{}' not found".format(python_exec))

    # store parameters
    with open(os.path.join(path_project, "parameters.txt"), "w", encoding="utf-8") as fd:
        fd.write("setup script = {}\n".format(setup_script))
        if bitstream_file:
            fd.write("bitstream file = {}\n".format(bitstream_file))
            fd.write("bitstream file SHA256 digest = {}\n".format(sha256_digest(bitstream_file)))
        fd.write("timebase = {}\n".format(timebase))
        fd.write("clock frequency [Hz] = {}\n".format(clock_frequency))
        fd.write("trigger threshold [mV] = {}\n".format(trigger_threshold))
        fd.write("number of measurements = {}\n".format(33))

    # start operation & device
    setup_cryptographic_operation(python_exec, setup_script, clock_frequency, bitstream_file)
    pico = start_picoscope()
    blocks_measured = 0  # how many blocks have been recorded?

    pathlib.Path(path_measurements).mkdir(parents=True, exist_ok=True)

    try:
        for i in range(measurements):
            # take test measurement
            block_crypto, block_trigger, _, _, interval_in_ns  = run_block_with_picoscope(pico, timebase=timebase)
            assert len(block_crypto) == len(block_trigger)
            #d = np.ndarray((2, len(block_crypto)), dtype=np.float16)
            #d[0, :] = block_crypto
            #d[1, :] = block_trigger
            write_mbin_file([block_crypto, block_trigger], os.path.join(path_project, "measurement{:03d}.mbin".format(i)))
            
            #print( timebase, timebase_to_sample_interval(timebase), interval_in_ns)
            #assert timebase_to_sample_interval(timebase) == interval_in_ns, interval_in_ns

            # determine high-ranges, high-clusters and the min/max of high-range lengths
            high_ranges = find_high_ranges(block_trigger, trigger_threshold)
            high_clusters = find_clusters_of_high_ranges(high_ranges)
            assert len(high_ranges) > 0
            assert len(high_clusters) > 0

            LOGGER.info("{} of {} measurements happened so far".format(i,measurements))

    finally:
        print("{} measurement blocks have been recorded".format(blocks_measured))
        stop_picoscope(pico)

    teardown_cryptographic_operation(python_exec, teardown_script)
    LOGGER.info("done at {}.".format(datetime.datetime.utcnow().isoformat()))


if __name__ == '__main__':
    loglevels = 'CRITICAL;DEBUG;ERROR;FATAL;INFO;NOTSET;WARN;WARNING'.split(';')

    parser = argparse.ArgumentParser(description=__doc__.strip())
    parser.add_argument('project', help='project name; <data-dir>/<project>/ directory is used for storing files')
    parser.add_argument('--data-dir', default='data', help='directory to store measurement results')
    parser.add_argument('--bitstream-file', help='the bitstream file to flash to device before setup')
    parser.add_argument('--setup-script', default='crypto_setup.py', help='the setup script to be used for configuring the cryptographic operation')
    parser.add_argument('--teardown-script', default='crypto_teardown.py', help='the teardown script run once the cryptographic operation finished')
    parser.add_argument('--timebase', default=8, type=int, help='timebase (configures sampling frequency) as defined by PicoScope 5203 Programmer Guide')
    parser.add_argument('--clock-frequency', default=10000000, type=int, help='PLL crypto clock frequency to be configured at CW305 in cycles per second')
    parser.add_argument('--measurements', default=10, type=int, help='how many times shall the cryptographic operation be recorded?')
    parser.add_argument('--trigger-threshold', default=200, type=int, help='trigger signal is below this threshold in mV? then discard power data')
    parser.add_argument('--python-exec', help='python3 executable to use to setup/teardown cryptographic operation')
    parser.add_argument('--log-level', dest='loglevel', default='DEBUG', choices=loglevels, help='loglevel for the logging module')
    parser.add_argument('--log-format', dest='logformat', default='%(asctime)s,%(levelname)s: %(message)s', help='log message for the logging module')

    args = parser.parse_args()
    main(args.data_dir, args.project, args.bitstream_file, args.setup_script, args.teardown_script, args.timebase, args.clock_frequency, args.measurements, args.trigger_threshold, args.python_exec)
